/* bidirectional search algorithms library
    Copyright (C) 2010 Thomas Schnattinger

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/ .
*/
/*! \file index_csa_fmi_text.hpp
    \brief Index backed by the occ function and samples of SA
	\author Thomas Schnattinger
*/
#ifndef INCLUDED_BIDIRECTIONAL_INDEX_CSA_FMI_TEXT
#define INCLUDED_BIDIRECTIONAL_INDEX_CSA_FMI_TEXT

#include <iostream>
#include <sdsl/vectors.hpp>
#include <sdsl/algorithms.hpp>
#include "indexes.hpp"
#include "wavelet_tree.hpp"

/*!
 * This class implements a compressed suffix array backed by the occ function.
 * So both forward and backward search are availible.
 */
template<class Occ = wavelet_tree<>, uint32_t SampleDens = 12>
class index_csa_fmi_text : public index {

	private:

		Occ m_occ;
		int_vector<> m_sa_sample; // suffix array samples
		int_vector<> text;

		inline size_t getSAValue(size_t i) const;

	public:

 		struct searchstate {
			size_t occ_begin;
			size_t occ_end;
			size_t length;
			searchstate() : occ_begin(0), occ_end(0), length(0) {}
			searchstate(size_t i, size_t j, size_t l) : occ_begin(i), occ_end(j), length(l) {}
			searchstate(const searchstate &s) : occ_begin(s.occ_begin), occ_end(s.occ_end), length(s.length) {}
		};

		//! Default Constructor
		index_csa_fmi_text() : index() {}

		/*! 
		 * Constructor for building the Index
		 * \param[in] str C-string of the text
		 */
		index_csa_fmi_text(const unsigned char *str) : index() {

			size_t n = strlen((const char*)str);
			int_vector<> sa(n+1, 0, bit_magic::l1BP(n+1)+1);
			algorithm::calculate_sa(str, n+1, sa);   // calculate the suffix array sa of str
			setText(str, n+1);

			text = int_vector<>(sa.size(), 0, bit_magic::l1BP(sigma)+1);
			for (size_t i=0; i<sa.size(); i++) text[i] = char2comp[str[i]];

			unsigned char *bwt = new unsigned char[n+1];
			{ /* Calculate Burrows-Wheeler-Transform */
				size_t i = 0;
				for(int_vector<>::const_iterator it = sa.begin(), end = sa.end(); it != end; ++it, ++i){
					bwt[i] = m_char2comp[str[(*it+n)%(n+1)]];
				}
			}
			m_occ = Occ(bwt, n+1, m_sigma);
			delete[] bwt;

			/* Construct the SA-Samples */
			m_sa_sample.setIntWidth( bit_magic::l1BP(sa.size())+1 );
			m_sa_sample.resize( (sa.size()+SampleDens-1)/SampleDens );
			size_t idx=0;
			size_t i=(sa.size()-1-SampleDens*(m_sa_sample.size()-1));
			for(int_vector<>::const_iterator it = sa.begin()+(ptrdiff_t)i; i < sa.size(); it += (ptrdiff_t)SampleDens, i += SampleDens, ++idx){
				m_sa_sample[idx] = *it;
			} 

		}

		/*!
		 * Initializes a searchstate to the searchstate of the empty word
		 * \param[out] s searchstate to be initialized
		 */
		void init_search_state(searchstate &s);

		/*!
		 * Search for character c in forward direction, starting with a given searchstate
		 * \param[in] c character to search for
		 * \param[in,out] s searchstate (will be overwritten)
		 * \return size of the new interval
		 */
		size_t forward_search(unsigned char c, searchstate &s);

		/*!
		 * Search for character c in backward direction, starting with a given searchstate
		 * \param[in] c character to search for
		 * \param[in,out] s searchstate (will be overwritten)
		 * \return size of the new interval
		 */
		size_t backward_search(unsigned char c, searchstate &s);

		/*!
		 * Extract the SA-values of the interval $[i..j]$
		 * \param[in] i begin of the interval (including)
		 * \param[in] j end of the interval (excluding)
		 * \param[out] occ result
		 */
		void extract_sa(size_t i, size_t j, int_vector<> &occ);

		/*!
		 * Serializes this data structure to an output stream
		 * \param[in,out] out output stream to write to
		 * \return number of bytes written to out
		 */
		size_t serialize(std::ostream &out) const;

		/*!
		 * Load this data structure from an input stream.
		 * \param[in,out] in input stream to read from
		 */
		void load(std::istream &in);

		/*!
		 * Calculate the used memory of this data structure by serializing into /dev/null
		 * \return size of this data structure in bytes
		 */
		size_t used_bytes() const;
};

template<class Occ, uint32_t SampleDens>
void index_csa_fmi_text<Occ, SampleDens>::init_search_state(searchstate& s) {
	s.occ_begin = 0;
	s.occ_end = size();
	s.length = 0;
}

template<class Occ, uint32_t SampleDens>
size_t index_csa_fmi_text<Occ, SampleDens>::forward_search(unsigned char c, searchstate &s) {
//	c = sa.char2comp[c];
	size_t begin = s.occ_begin, end = s.occ_end, mid; 
	while(begin != end){
		mid = (begin+end)>>1;  // begin <= mid < end
		if( text[getSAValue(mid)+s.length] < c ){
			begin = mid+1;
		}else{
			end = mid; 
		}    
	}    
	s.occ_begin = begin;
	end = s.occ_end;
	if(begin >= end){
		return s.occ_begin = s.occ_end = 0; 
	}    

	while(begin != end){
		mid = (begin+end)>>1; // begin <= mid < end
		if( text[getSAValue(mid)+s.length] <= c ){
			begin = mid+1;  
		}else{
			end = mid; 
		}    
	}    
	s.occ_end = end; 
	if( s.occ_end <= s.occ_begin ){
		return s.occ_begin = s.occ_end = 0; 
	}    

	s.length += 1;	
	return s.occ_end-s.occ_begin;
}

template<class Occ, uint32_t SampleDens>
size_t index_csa_fmi_text<Occ, SampleDens>::backward_search(unsigned char c, searchstate &s) {
//	c = sa.char2comp[c];

	s.length += 1;
	s.occ_begin = C[c] + m_occ.occ(c, s.occ_begin);
	s.occ_end = C[c] + m_occ.occ(c, s.occ_end);
	
	return s.occ_end-s.occ_begin;
}


template<class Occ, uint32_t SampleDens>
void index_csa_fmi_text<Occ, SampleDens>::extract_sa(size_t occ_begin, size_t occ_end, int_vector<> &occs) {
	occs.resize(occ_end-occ_begin);
	for (size_t k=0; k<occ_end-occ_begin; k++) occs[k] = getSAValue(occ_begin+k);
}

template<class Occ, uint32_t SampleDens>
inline size_t index_csa_fmi_text<Occ, SampleDens>::getSAValue(size_t i)const{
//	cout << "operator[] (" << i << ")... size()="<<size() << endl;
	size_t off = 0;
	while( (size()-1-i) % SampleDens != 0 ){// while SA[i] is not sampled
		// go to the position where SA[i]-1 is located
		// i := LF(i)
		size_t bwt_result = m_occ.extract(i);
		i = C[bwt_result] + m_occ.occ(bwt_result, i);
		// add 1 to the offset
		++off;
	}
	size_t result = m_sa_sample[i / SampleDens];
//	cout << "result of operator[] ("<<i<<") = " << (result+off)%size()<<endl;
	return (result+off) % size();
}

template<class Occ, uint32_t SampleDens>
size_t index_csa_fmi_text<Occ, SampleDens>::serialize(std::ostream &out)const{
	size_t written_bytes = 0;
	out.put('5');
	written_bytes += sizeof(char);
	written_bytes += superserialize(out);
	written_bytes += m_occ.serialize(out);
	written_bytes += m_sa_sample.serialize(out);
	written_bytes += text.serialize(out);
	return written_bytes;
}

template<class Occ, uint32_t SampleDens>
void index_csa_fmi_text<Occ, SampleDens>::load(std::istream &in) {
	if (in.get()!='5') {
		std::cerr << "wrong index!!";
		throw("wrong index!!");
	}
	superload(in);
	m_occ.load(in);
	m_sa_sample.load(in);
	text.load(in);	
}

template<class Occ, uint32_t SampleDens>
size_t index_csa_fmi_text<Occ, SampleDens>::used_bytes() const {
	boost::onullstream dev_null;
   	return serialize(dev_null);
}   

#endif
