/* bidirectional search algorithms library
    Copyright (C) 2010 Thomas Schnattinger

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/ .
*/
/*! \file index_sa_isa.hpp
    \brief Index consisting of SA and inverse SA
	\author Thomas Schnattinger
*/
#ifndef INCLUDED_BIDIRECTIONAL_INDEX_SA_ISA
#define INCLUDED_BIDIRECTIONAL_INDEX_SA_ISA

#include <iostream>
#include <boost/test/utils/nullstream.hpp>
#include <sdsl/int_vector.hpp>
#include "indexes.hpp"

/*! 
 * This index uses binary search on the suffix array for forward search 
 * and binary search on the psi function calculated by psi(i) = sa[isa[i]-1]]
 */
class index_sa_isa : public index {
	private:
		inline size_t psi(size_t i) const;
		uint8_t lookup_C(size_t i) const;

		int_vector<> sa;
		int_vector<> isa;

	public:

		struct searchstate {
			size_t occ_begin;
			size_t occ_end;
			size_t length;
			searchstate() : occ_begin(0), occ_end(0), length(0) {}
			searchstate(size_t i, size_t j, size_t l) : occ_begin(i), occ_end(j), length(l) {}
			searchstate(const searchstate &s) : occ_begin(s.occ_begin), occ_end(s.occ_end), length(s.length) {}
		};

		//! Default constructor
		index_sa_isa() : index() {}

		/*! 
		 * Constructor for building the Index
		 * \param[in] str C-string of the text
		 */
		index_sa_isa(const unsigned char* str) : index() {
			size_t n = strlen((const char*)str);
			setText(str, n+1);
			sa = int_vector<>(n+1, 0, bit_magic::l1BP(n+1)+1);
			algorithm::calculate_sa(str, n+1, sa);	 // calculate the suffix array sa of str

			isa = sa;
			for(size_t i=0; i<sa.size(); ++i){
				isa[sa[i]] = i;
			}
		}

		/*!
		 * Initializes a searchstate to the searchstate of the empty word
		 * \param[out] s searchstate to be initialized
		 */
		void init_search_state(searchstate &s);

		/*!
		 * Search for character c in forward direction, starting with a given searchstate
		 * \param[in] c character to search for
		 * \param[in,out] s searchstate (will be overwritten)
		 * \return size of the new interval
		 */
		size_t forward_search(unsigned char c, searchstate &s);

		/*!
		 * Search for character c in backward direction, starting with a given searchstate
		 * \param[in] c character to search for
		 * \param[in,out] s searchstate (will be overwritten)
		 * \return size of the new interval
		 */
		size_t backward_search(unsigned char c, searchstate &s);

		/*!
		 * Extract the SA-values of the interval $[i..j]$
		 * \param[in] i begin of the interval (including)
		 * \param[in] j end of the interval (excluding)
		 * \param[out] occ result
		 */
		void extract_sa(size_t i, size_t j, int_vector<> &occ);

		/*!
		 * Serializes this data structure to an output stream
		 * \param[in,out] out output stream to write to
		 * \return number of bytes written to out
		 */
		size_t serialize(std::ostream &out) const;

		/*!
		 * Load this data structure from an input stream.
		 * \param[in,out] in input stream to read from
		 */
		void load(std::istream &in);

		/*!
		 * Calculate the used memory of this data structure by serializing into /dev/null
		 * \return size of this data structure in bytes
		 */
		size_t used_bytes() const;
};

inline size_t index_sa_isa::psi(size_t i) const {
	return isa[(sa[i]+1)%size()];
}

uint8_t index_sa_isa::lookup_C(size_t i) const {
	uint8_t a=0, b=sigma, m;
	while (a+1!=b) {
		m = (a + b) >> 1;
		if (C[m]<=i) {
			if (C[m+1]<=i) {
				a = m;
			} else {
				return m;
			}   
		} else {
			b = m;
		}   
	}   
	return a;
}

void index_sa_isa::init_search_state(searchstate& s) {
	s.occ_begin = 0;
	s.occ_end = size();
	s.length = 0;
}

size_t index_sa_isa::forward_search(unsigned char c, searchstate &s) {
//	c = sa.char2comp[c];
	size_t begin = s.occ_begin, end = s.occ_end, mid; 
	while(begin != end){
		mid = (begin+end)>>1;  // begin <= mid < end
		if( lookup_C(isa[sa[mid]+s.length]) < c ){
			begin = mid+1;
		}else{
			end = mid; 
		}    
	}    
	s.occ_begin = begin;
	end = s.occ_end;
	if(begin >= end){
		return s.occ_begin = s.occ_end = 0; 
	}    

	while(begin != end){
		mid = (begin+end)>>1; // begin <= mid < end
		if( lookup_C(isa[sa[mid]+s.length]) <= c ){
			begin = mid+1;  
		}else{
			end = mid; 
		}    
	}    
	s.occ_end = end; 
	if( s.occ_end <= s.occ_begin ){
		return s.occ_begin = s.occ_end = 0; 
	}    
	
	s.length += 1;
	return s.occ_end-s.occ_begin;
}

size_t index_sa_isa::backward_search(unsigned char c, searchstate &s) {
//	c = sa.char2comp[c];
	size_t c_begin = C[c];  // interval begin of character c (inclusive)
	size_t c_end   = C[c+1];// interval end of character c (exclusive)
	if( c_begin == c_end ){
		s.occ_begin = s.occ_end = c_begin;
		return 0;
	}

	size_t begin = c_begin, end = c_end, mid; 
	while(begin != end){
		mid = (begin+end)>>1;  // begin <= mid < end
		if( psi(mid) < s.occ_begin ){
			begin = mid+1;
		}else{
			end = mid; 
		}    
	}    
	s.occ_begin = begin;
	if(begin < c_end){
		s.occ_begin = begin;
	}else{
		return s.occ_begin = s.occ_end = 0; 
	}    
	end = c_end;
	if(begin >= end){
		return s.occ_begin = s.occ_end = 0; 
	}    

	while(begin != end){
		mid = (begin+end)>>1; // begin <= mid < end
		if( psi(mid) < s.occ_end ){
			begin = mid+1;  
		}else{
			end = mid; 
		}
	}    
	s.occ_end = end; 
	if( s.occ_end <= s.occ_begin ){
		return s.occ_begin = s.occ_end = 0; 
	}    
	if( end > c_begin){
		s.occ_end = end; 
	}else{
		return s.occ_end = s.occ_begin = 0; 
	}    
	
	s.length += 1;
	return s.occ_end-s.occ_begin;
}

void index_sa_isa::extract_sa(size_t occ_begin, size_t occ_end, int_vector<> &occ) {
	occ.resize(occ_end-occ_begin);
	for (size_t k=0; k<occ_end-occ_begin; k++) occ[k] = sa[occ_begin+k];
}

/*void index_sa_isa::extract_text(int start, int length, unsigned char* str) {
	size_t k = isa[start];
	for (size_t i=0; i<length; i++) {
		str[i] = comp2char[lookup_C(k)];
		k = psi(k);
	}
	str[length] = '\0';
}*/

size_t index_sa_isa::serialize(std::ostream &out) const {
	size_t written_bytes = 0;
	out.put('2');
	written_bytes += sizeof(char);
	written_bytes += index::superserialize(out);
	written_bytes += sa.serialize(out);
	written_bytes += isa.serialize(out);
	return written_bytes;
}

void index_sa_isa::load(std::istream &in) {
	if (in.get()!='2') {
		std::cerr << "wrong index!!";
		throw("wrong index!!");
	}
	superload(in);
	sa.load(in);
	isa.load(in);
}

size_t index_sa_isa::used_bytes() const {
	boost::onullstream dev_null;
	return serialize(dev_null);
}

#endif
