/* bidirectional search algorithms library
    Copyright (C) 2010 Thomas Schnattinger

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/ .
*/
/*! \file index_sa_text_psi.hpp
    \brief Index using SA, Text and Psi function
	\author Thomas Schnattinger
*/
#ifndef INCLUDED_BIDIRECTIONAL_INDEX_SA_TEXT_PSI
#define INCLUDED_BIDIRECTIONAL_INDEX_SA_TEXT_PSI

#include <iostream>
#include <boost/test/utils/nullstream.hpp>
#include <sdsl/vectors.hpp>
#include <sdsl/algorithms.hpp>
#include "indexes.hpp"

/*!
 * This index uses binary search on the suffix array for forward search
 * and the psi function (which is explicitly stored) for backward search
 */
class index_sa_text_psi : public index {
	private:
		int_vector<> sa;
		int_vector<> text;
		int_vector<> psi;

	public:

		struct searchstate {
			size_t occ_begin;
			size_t occ_end;
			size_t length;
			searchstate() : occ_begin(0), occ_end(0), length(0) {}
			searchstate(size_t i, size_t j, size_t l) : occ_begin(i), occ_end(j), length(l) {}
			searchstate(const searchstate &s) : occ_begin(s.occ_begin), occ_end(s.occ_end), length(s.length) {}
		};

		//! Default constructor
		index_sa_text_psi() : index() {}

		/*! 
		 * Constructor for building the Index
		 * \param[in] str C-string of the text
		 */
		index_sa_text_psi(const unsigned char* str) : index() {

			size_t n = strlen((const char*)str);
			sa = int_vector<>(n+1, 0, bit_magic::l1BP(n+1)+1);
			algorithm::calculate_sa(str, n+1, sa);   // calculate the suffix array sa of str
			sdsl::algorithm::sa2psi(sa, psi);
			setText(str, n+1);

			text = int_vector<>(sa.size(), 0, bit_magic::l1BP(sigma)+1);
			for (size_t i=0; i<sa.size(); i++) text[i] = char2comp[str[i]];
		}

		/*!
		 * Initializes a searchstate to the searchstate of the empty word
		 * \param[out] s searchstate to be initialized
		 */
		void init_search_state(searchstate &s);

		/*!
		 * Search for character c in forward direction, starting with a given searchstate
		 * \param[in] c character to search for
		 * \param[in,out] s searchstate (will be overwritten)
		 * \return size of the new interval
		 */
		size_t forward_search(unsigned char c, searchstate &s);

		/*!
		 * Search for character c in backward direction, starting with a given searchstate
		 * \param[in] c character to search for
		 * \param[in,out] s searchstate (will be overwritten)
		 * \return size of the new interval
		 */
		size_t backward_search(unsigned char c, searchstate &s);

		/*!
		 * Extract the SA-values of the interval $[i..j]$
		 * \param[in] i begin of the interval (including)
		 * \param[in] j end of the interval (excluding)
		 * \param[out] occ result
		 */
		void extract_sa(size_t i, size_t j, int_vector<> &occ);

		/*!
		 * Serializes this data structure to an output stream
		 * \param[in,out] out output stream to write to
		 * \return number of bytes written to out
		 */
		size_t serialize(std::ostream &out) const;

		/*!
		 * Load this data structure from an input stream.
		 * \param[in,out] in input stream to read from
		 */
		void load(std::istream &in);

		/*!
		 * Calculate the used memory of this data structure by serializing into /dev/null
		 * \return size of this data structure in bytes
		 */
		size_t used_bytes() const;
};	

void index_sa_text_psi::init_search_state(searchstate& s) {
	s.occ_begin = 0;
	s.occ_end = size();
	s.length = 0;
}

size_t index_sa_text_psi::forward_search(unsigned char c, searchstate &s) {
//	c = sa.char2comp[c];
	size_t begin = s.occ_begin, end = s.occ_end, mid; 
	while(begin != end){
		mid = (begin+end)>>1;  // begin <= mid < end
		if( text[sa[mid]+s.length] < c ){
			begin = mid+1;
		}else{
			end = mid; 
		}    
	}    
	s.occ_begin = begin;
	end = s.occ_end;
	if(begin >= end){
		return s.occ_begin = s.occ_end = 0; 
	}    

	while(begin != end){
		mid = (begin+end)>>1; // begin <= mid < end
		if( text[sa[mid]+s.length] <= c ){
			begin = mid+1;  
		}else{
			end = mid; 
		}    
	}    
	s.occ_end = end; 
	if( s.occ_end <= s.occ_begin ){
		return s.occ_begin = s.occ_end = 0; 
	}    
	
	s.length += 1;
	return s.occ_end-s.occ_begin;
}

size_t index_sa_text_psi::backward_search(unsigned char c, searchstate &s) {
//	c = sa.char2comp[c];
	size_t c_begin = C[c];  // interval begin of character c (inclusive)
	size_t c_end   = C[c+1];// interval end of character c (exclusive)
	if( c_begin == c_end ){
		s.occ_begin = s.occ_end = c_begin;
		return 0;
	}

	size_t begin = c_begin, end = c_end, mid; 
	while(begin != end){
		mid = (begin+end)>>1;  // begin <= mid < end
		if( psi[mid] < s.occ_begin ){
			begin = mid+1;
		}else{
			end = mid; 
		}    
	}    
	s.occ_begin = begin;
	if(begin < c_end){
		s.occ_begin = begin;
	}else{
		return s.occ_begin = s.occ_end = 0; 
	}    
	end = c_end;
	if(begin >= end){
		return s.occ_begin = s.occ_end = 0; 
	}    

	while(begin != end){
		mid = (begin+end)>>1; // begin <= mid < end
		if( psi[mid] < s.occ_end ){
			begin = mid+1;  
		}else{
			end = mid; 
		}
	}    
	s.occ_end = end; 
	if( s.occ_end <= s.occ_begin ){
		return s.occ_begin = s.occ_end = 0; 
	}    
	if( end > c_begin){
		s.occ_end = end; 
	}else{
		return s.occ_end = s.occ_begin = 0; 
	}    
	
	s.length += 1;
	return s.occ_end-s.occ_begin;
}

void index_sa_text_psi::extract_sa(size_t occ_begin, size_t occ_end, int_vector<> &occ) {
	occ.resize(occ_end-occ_begin);
	for (size_t k=0; k<occ_end-occ_begin; k++) occ[k] = sa[occ_begin+k];
}

/*void index_sa_text_psi::extract_text(int start, int length, unsigned char* str) {
	for (size_t i=0; i<length; i++) {
		str[i] = comp2char[text[start+i]];
	}
	str[length] = '\0';
}*/

size_t index_sa_text_psi::serialize(std::ostream &out) const {
	size_t written_bytes = 0;
	out.put('1');
	written_bytes += sizeof(char);
	superserialize(out);
	written_bytes += sa.serialize(out);
	written_bytes += text.serialize(out);
	written_bytes += psi.serialize(out);
	return written_bytes;
}

void index_sa_text_psi::load(std::istream &in) {
	if (in.get()!='1') {
		std::cerr << "wrong index!!";
		throw("wrong index!!");
	}
	superload(in);
	sa.load(in);
	text.load(in);
	psi.load(in);
}

size_t index_sa_text_psi::used_bytes() const {
	boost::onullstream dev_null;
	return serialize(dev_null);
}

#endif
