/* bidirectional search algorithms library
    Copyright (C) 2010 Thomas Schnattinger

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/ .
*/
/*! \file starttestcases.cpp
    \brief Run the testcase mentioned in the thesis on a given index
	\author Thomas Schnattinger
*/
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include "indexes.hpp"
#include "bidirectional.hpp"
#include <sdsl/testutils.hpp>
#include <sdsl/coder.hpp>

using namespace std;
using namespace sdsl;

void printbignumber(uint64_t n) {
	if (n>=1000) {
		printbignumber(n/1000);
		printf(".");
		printf("%03d", n%1000);
	} else printf("%d", n%1000);
}

template <class Index>
void test_bidirectional(char* indexfile, uint32_t durchgaenge) {
	typedef typename Index::searchstate searchstate;
	typedef size_t (*testpattern)(Index&, std::ostream&);
	typedef size_t (*testpattern_with_parameter)(Index&, size_t, std::ostream&);
	size_t hits;

	testpattern tests1[] = {
		&locate_testpattern_hairpin1,
		&locate_testpattern_hairpin2,
		&locate_testpattern_hairpin4,
		//&locate_testpattern_sequence,
		NULL
	};

	testpattern_with_parameter tests2[] = {
		&locate_testpattern_hloop,
		//&locate_testpattern_hloop,
		&locate_testpattern_acloop,
		&locate_testpattern_acloop,
		&locate_testpattern_acloop,
		NULL
	};
	size_t tests2_parameter[] = {
		5, 
//		10, 
		5, 
		10, 
		15
	};

	char outputfilename[100];

	Index index;
	ifstream in(indexfile);
	index.load(in);
	in.close();

	printf("Speicherplatz: ");
	printbignumber(index.used_bytes());
	printf(" Bytes\n");

	stop_watch stopwatch;
	for (uint32_t durchgang=0; durchgang<durchgaenge; durchgang++) {

		sprintf(outputfilename, "out_%02d", durchgang);
		std::ofstream out(outputfilename);

		for (int i=0; tests1[i]!=NULL; i++) {
			stopwatch.start();
			hits = tests1[i](index, out);
			stopwatch.stop();
			printf("  Muster %d mal gefunden...\n", hits);
			printf("  Zeit zum Suchen: %d ms.\n", (int)stopwatch.getUserTime());
		}
	
		for (int i=0; tests2[i]!=NULL; i++) {
			stopwatch.start();
			hits = tests2[i](index, tests2_parameter[i], out);
			stopwatch.stop();
			printf("  Muster %d mal gefunden...\n", hits);
			printf("  Zeit zum Suchen: %d ms.\n", (int)stopwatch.getUserTime());
		}
	
		out.close();
	}
}

int main(int argc, char** argv) {

	if (argc!=2 && argc!=3) {
		printf("\n");
		printf("Starts the test cases mentioned in the thesis.\n");
		printf("\n");
		printf("usage: %s indexfile [durchgaenge (default=1)]\n", argv[0]);
		printf("\n");
		return 1;
	}

	char* indexfile = argv[1];

	ifstream in(indexfile);
	char index = in.get();
	in.close();

	int durchgaenge = argc==2 ? 1 : atoi(argv[2]);

	switch (index) {
		case '1':
			cout << "Found index_sa_text_psi" << endl;
			test_bidirectional<index_sa_text_psi>(indexfile, durchgaenge);
			break;
		case '2':
			cout << "Found index_sa_isa" << endl;
			test_bidirectional<index_sa_isa>(indexfile, durchgaenge);
			break;
		case '3':
			cout << "Found index_sa_text_occ<>" << endl;
			test_bidirectional<index_sa_text_occ<> >(indexfile, durchgaenge);
			break;
		case '4':
			cout << "Found index_csa_psi_text<>" << endl;
			test_bidirectional<index_csa_psi_text<> >(indexfile, durchgaenge);
			break;
		case '5':
			cout << "Found index_csa_fmi_text<>" << endl;
			test_bidirectional<index_csa_fmi_text<> >(indexfile, durchgaenge);
			break;
		case '6':
			cout << "Found index_bidirectional_waveletindex<>" << endl;
			test_bidirectional<index_bidirectional_waveletindex<> >(indexfile, durchgaenge);
			break;
		default:
			cerr << "ERROR. wrong index number." << endl;
			return 1;
	}

	return 0;
}
