#include "sdsl/suffixtrees.hpp"
#include "sdsl/algorithms.hpp"

#include <iostream>
#include <vector>
#include <list>
#include <string>
#include <utility>
#include <iomanip>

using namespace sdsl;
using namespace std;

typedef cst_sct<> tCST;
typedef tCST::size_type size_type;
typedef vector<size_type> tVI;
typedef pair<size_type, size_type> tPII;
typedef vector<tPII> tVPII;


template<class EncVector, uint32_t SampleDens, uint32_t InvSampleDens, uint8_t fixedIntWidth>
typename csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::size_type backward_search(
			const csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth> &csa, 	
			const unsigned char c,
			typename csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::size_type &lb,
			typename csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::size_type &rb
)
{
	typedef typename csa_sada_prac<EncVector, SampleDens, InvSampleDens, fixedIntWidth>::size_type size_type;
		size_type cc = csa.char2comp[c];
		if(cc==0){
			lb=rb=0;
			return 0;
		}
		size_type c_begin = csa.C[cc];  // interval begin of character c (inclusive)
		size_type c_end   = csa.C[cc+1];// interval end of character c (exclusive)
		if( c_begin == c_end ){
			lb = rb = 0;
			return 0;
		}
		// logarithmic solution
		size_type begin = c_begin, end = c_end, mid;
		while(begin != end){
			mid = (begin+end)>>1;  // begin <= mid < end
			if( csa.psi[mid] < lb ){
				begin = mid+1;
			}else{
				end = mid;
			}
		}
		assert(begin==c_end or csa.psi[begin] >= lb);
		if(begin < c_end){
			lb = begin;
		}else{
			return lb = rb = 0;
		}
		end = c_end;
		while(begin != end){
			mid = (begin+end)>>1; // begin <= mid < end
			if( csa.psi[mid] < rb ){
				begin = mid+1;	
			}else{
				end = mid;
			}
		}
		assert(end==c_end or csa.psi[end] >= rb);
		assert(end == c_begin or csa.psi[end-1]<rb);
		if( end > c_begin){
			rb = end;
		}else{
			return rb = lb = 0;
		}
	return rb-lb;
}


//! Calculate the matching statistics of string s2 against the suffix tree cst1 of string s1.
/*! \param cst1 The (compressed) suffix tree of s1 supporting the parent operation.
 *	\param s2	String s2 for which we calculate for each position i the longest common prefix of s2[i..n2] and a suffix of s1.
 *	\param n2	Length of string s2.
 *	\param ms	Reference to the result. Position i contains the length of the longeset common prefix of s2[i..n2] and a suffix of s1.
 *	\param ms_intervals	The \f$\omega\f$-interval ...TODO
 *  
 *	\par Time complexity
 *	\f$ \Order{n_2 \log|n| } \f$
 */
void matching_statistics(const tCST &cst1, const unsigned char *s2, size_type n2, tVI &ms, tVPII &ms_intervals){
	typedef tCST::node_type node_type;
	ms.resize(n2); ms_intervals.resize(n2);
	size_type p2 = n2;
	size_type i = 0, j = cst1.size();
	size_type c = 0;
	while( p2 > 0 ){
		size_type lb=i, rb=j;
		if( backward_search(cst1.csa, s2[p2-1], lb, rb) ){
			++c;
			ms[p2-1] = c; ms_intervals[p2-1] = tPII(lb, rb-1);
			i = lb; j = rb;
			--p2;
		}else if( 0==c ){
			ms[p2-1] = 0; ms_intervals[p2-1] = tPII(0, cst1.size());
			--p2;
		}else{
			node_type p = cst1.parent( node_type(c, i, j-1) );
			c = cst1.depth(p);
			i = cst1.leftmost_suffix_array_index_in_the_subtree(p); 
			j = cst1.rightmost_suffix_array_index_in_the_subtree(p)+1;
		}
	}
}

struct path_item{
	tCST::size_type c, lb, rb, p;
	path_item(size_type c=0, size_type lb=0, size_type rb=0, size_type p=0):c(c),lb(lb),rb(rb),p(p){}
};

void report_maximal_exact_match(const tCST &cst1, const unsigned char *s2, size_type c_, size_type p2_, size_type k){
	size_type sa_k = cst1.csa[k];
	if( p2_ == 0){
		cout<<"("<<c_<<","<<sa_k+1<<","<<p2_+1<<")"<<endl;
	}else{ // p2_ > 0
		size_type lf_k = cst1.csa( (sa_k-1+cst1.csa.size()) % cst1.csa.size() );
		size_type cc2 = cst1.csa.char2comp[s2[p2_-1]];
		if( !(cst1.csa.C[cc2] <= lf_k and lf_k < cst1.csa.C[cc2+1]) ){
			cout<<"("<<c_<<","<<sa_k+1<<","<<p2_+1<<")"<<endl;
		}
	}
}

//! Calculate all maximal exact matches of string s1 and s2 with minimal length l.
/*!
 *	\param cst1 The (compressed) suffix tree of s1 supporting the parent operation.
 *  \param s2  	String s2 for which we calculate all maximal exact matches with string s1.
 *	\param n2	The length of s2.
 *	\param l	Threshold value for the length of the minimal maximal matches. \f$ l > 0 \f$
 */
void maximal_exact_matches(const tCST &cst1, const unsigned char *s2, size_type n2, tCST::size_type l){
	typedef tCST::node_type node_type;
	typedef tCST::size_type	size_type;
	typedef list<path_item> path_type;
	assert( l > 0 );
	
	size_type p2 = n2;
	size_type i = 0, j = cst1.size();
	size_type c = 0;
	while( p2 > 0 ){
		path_type path;
		size_type lb = i, rb = j;
		backward_search(cst1.csa, s2[p2-1], lb, rb);	
		while(lb != rb and p2 > 0){
			++c;
			if( c >= l ){
				path.push_back( path_item(c, lb, rb, p2-1) );
			}
			i = lb; j = rb;
			--p2;
			backward_search(cst1.csa, s2[p2-1], lb, rb);
		}
		for(path_type::const_iterator it = path.begin(); it!=path.end(); ++it){
			size_type c_ 	= it->c;
			size_type lb_ 	= it->lb, rb_ = it->rb;
			lb = lb_; rb = lb_;
			size_type p2_	= it->p;
			while( c_ >= l  ){
				for(size_type k = lb_; k < lb; ++k){
					report_maximal_exact_match(cst1, s2, c_, p2_, k);
				}
				for(size_type k = rb; k < rb_; ++k){
					report_maximal_exact_match(cst1, s2, c_, p2_, k);
				}
				lb = lb_; rb = rb_;
				node_type p = cst1.parent( node_type(c_, lb_, rb_-1) );
				c_ 			= cst1.depth(p);
				lb_			= cst1.leftmost_suffix_array_index_in_the_subtree(p);
				rb_			= cst1.rightmost_suffix_array_index_in_the_subtree(p)+1;
			}
		}
		if( 0==c ){
			--p2;
		}else{
			node_type p = cst1.parent( node_type(c, i, j-1) );
			c 			= cst1.depth(p);
			i		 	= cst1.leftmost_suffix_array_index_in_the_subtree(p); 
			j 			= cst1.rightmost_suffix_array_index_in_the_subtree(p)+1;
		}
	}
}

int main(){
	string s1, s2;
	cin>>s1>>s2;
	// create compresssed suffix tree
	tCST cst1((const unsigned char*)s1.c_str()); 

	for(size_type i=0; i<cst1.csa.size(); ++i)
	{
		cout<<setw(3)<<i<<" "<<setw(3)<<cst1.csa[i]<<" "<<setw(3)<<cst1.csa.psi[i]<<" "<<setw(3)<<cst1.lcp[i]<<" "<< s1.c_str()+cst1.csa[i] <<endl;
	}

	cout<< cst1.root() << endl;
	tVI ms;
	tVPII ms_intervals;
	matching_statistics(cst1, (const unsigned char*)s2.c_str(), s2.size(), ms, ms_intervals);
	for(size_t i=0; i<ms.size(); ++i)
	{
		cout<<i<<" "<<ms[i]<<" ["<<ms_intervals[i].first+1<<".."<<ms_intervals[i].second+1<<"]"<<endl;
	}

	maximal_exact_matches(cst1, (const unsigned char*)s2.c_str(), s2.size(), 2);

	// iterator through the tree in dfs order 
/*	for(tCST::iterator it = cst1.begin(); it!=cst1.end(); ++it ){
		cout<< *it << " " << (int)it.visit() <<endl;
	}
*/	
}
